"""Parts module.

This module allows PyEnSight to control the parts in the EnSight session.

Example for selecting all 3D parts:

(PyEnSight)
>>> from ansys.pyensight.core import LocalLauncher
>>> session = LocalLauncher().start()
>>> parts = session.ensight.utils.parts
>>> parts.select_by_dimension(3)

(EnSight)
>>> from ensight.utils import parts
>>> parts.select_by_dimension(3)

"""

from typing import TYPE_CHECKING, Dict, Optional, Union

try:
    import ensight
    from ensight.objs import ensobjlist  # type: ignore
except ImportError:
    from ansys.pyensight.core.listobj import ensobjlist

if TYPE_CHECKING:
    try:
        from ensight.objs import ENS_PART  # type: ignore
    except ImportError:
        from ansys.api.pyensight import ensight_api
        from ansys.api.pyensight.ens_part import ENS_PART


class Parts:
    """Controls the parts in the current EnSight ``Session`` instance."""

    def __init__(self, ensight: Union["ensight_api.ensight", "ensight"]):
        self.ensight = ensight

    def select_parts_by_dimension(self, dimension: int) -> ensobjlist["ENS_PART"]:
        """Select parts by the input dimension and return the parts found.

        Parameters
        ----------
        dimension : int
            Dimension for selecting parts.

        Returns
        -------
        ensobjlist["ENS_PART"]
            found (ensobjlist): List of parts found.

        """
        parts = self.ensight.objs.core.PARTS
        parts.set_attr("SELECTED", False)
        found = parts.find(True, f"HAS{dimension}DELEMENTS")
        found.set_attr("SELECTED", True)
        return found

    def select_parts_invert(self) -> ensobjlist["ENS_PART"]:
        """Select parts currently not selected, deselecting the previously selected parts.

        Returns
        -------
        ensobjlist["ENS_PART"]
            Updated list of parts selected.

        """
        self.ensight.part.select_invert()
        parts = self.ensight.objs.core.PARTS
        return parts.find(True, "SELECTED")

    def select_parts_by_tag(
        self,
        tag: Optional[str] = None,
        value: Optional[str] = None,
        tagdict: Optional[Dict[str, str]] = None,
    ) -> ensobjlist["ENS_PART"]:
        """Select parts by the input dimension and return the parts found.

        Parameters
        ----------
        tag : str, optional
            Tag for finding the parts.
        value : str, optional
            Value for finding the parts.
        tagdict : dict, optional
            Dictionary containing the key and value pairs for finding
            the parts. Only the parts that have all the keys and corresponding
            values are returned. If a value for this parameter is supplied, it
            takes precedence over the valeus supplied for the ``tag`` and
            ``value`` parameters.

        Returns
        -------
        ensobjlist["ENS_PART"]
            List of parts found. If no arguments are given, all parts are returned.

        """
        parts = self.ensight.objs.core.PARTS
        metadata = {p: p.METADATA for p in parts}
        found = ensobjlist()
        if not tag and not value and not tagdict:
            self.ensight.part.select_all()
            return parts
        if not tagdict:
            if tag and value:
                found = ensobjlist([p for p, met in metadata.items() if met.get(tag) == value])
            elif value and not tag:
                found = ensobjlist([p for p, met in metadata.items() if value in met.values()])
            elif tag and not value:
                found = ensobjlist([p for p, met in metadata.items() if tag in met.keys()])
        else:
            found = ensobjlist(
                [
                    p
                    for p, met in metadata.items()
                    if all(met.get(k) == v for k, v in tagdict.items())
                ]
            )
        if found:
            found.set_attr("SELECTED", True)
        return found



"""EnsContext module

Interface to objects that contain a representation of an EnSight session state.
Three types of content are supported:

    #. Full context: a complete EnSight context file.
    #. Simple context: an EnSight context file without any data loading reference.

"""
import base64
import io
import os
import tempfile
from typing import Any, Optional, Union
import warnings
import zipfile


class EnsContext:
    """A saved EnSight session state

    This object allows for the generation and application of the
    EnSight "state".  The object may store a EnSight "context",
    or a "data-less context" representation.  The object can
    save() and load() the context to/from disk.

    Parameters
    ----------
    filename : str, optional
        If specified, "load()" the named context file after
        creating the object instance.

    """

    _UNKNOWN: int = 0
    _FULL_CONTEXT: int = 1
    _SIMPLE_CONTEXT: int = 2

    def __init__(self, filename: Optional[str] = None) -> None:
        """Initialized EnsContext."""
        self._type: int = self._UNKNOWN
        self._buffer: io.BytesIO = io.BytesIO()
        if filename is not None:
            self.load(filename)

    def _set_type(self, names: list) -> None:
        """Update the 'type' of the context

        Look though the file files stored in the zip file.  Look for the special
        embedded "type" files and set the object type accordingly.

        Parameters
        ----------
        names: list
            A list of filenames in the zip file.

        """
        self._type = self._UNKNOWN
        if "fullcontext.txt" in names:
            self._type = self._FULL_CONTEXT
        elif "simplecontext.txt" in names:
            self._type = self._SIMPLE_CONTEXT

    def load(self, filename: str) -> None:
        """Read a context from a local zip file

        Given the name of a context file, read it into memory and make it available
        for use by the PyEnSight Session methods.

        Parameters
        ----------
        filename: str
            The name of the file to read.

        """
        if not zipfile.is_zipfile(filename):
            raise RuntimeError(f"'{filename}' is not a saved context file.")
        with open(filename, "rb") as f:
            data = f.read()
        self._from_data(data)

    def _from_data(self, data: Union[bytes, str]) -> None:
        """Read a context from a blob or string

        Given a context file in the form of a bytes object or a
        the same bytes object encoded into a string using base64
        encoding.

        Parameters
        ----------
        data: Union[bytes, str]
            A bytes or string object of the contents of a
            context zip file.

        """
        if type(data) != bytes:
            data = base64.b64decode(data)
        self._buffer = io.BytesIO(data)
        the_file = zipfile.ZipFile(self._buffer, "r")
        self._set_type(the_file.namelist())

    def save(self, filename: str) -> None:
        """Save the context information to a file

        Save the current context to disk.

        Parameters
        ----------
        filename: str
            Name of the file to save.

        """
        data = self._buffer.getvalue()
        if len(data) < 1:
            raise RuntimeError("No context data to save")
        with open(filename, "wb") as fp:
            fp.write(data)

    def _data(self, b64: bool = False) -> Union[bytes, str]:
        """Return a representation of the context file as a string or bytes object

        Either a bytes object or a string (base64 encoded bytes object)
        representation of the current context file is returned.

        Parameters
        ----------
        b64: bool
            If True, return the bytes representation encoded into a string
            object using base64 encoding. By default, false.

        Returns
        -------
        Union[bytes, str]
            A bytes object or a string object.

        """
        data = self._buffer.getvalue()
        if b64:
            return base64.b64encode(data).decode("ascii")
        return data

    def _build_from_directory(self, pathname: str) -> None:
        """Create a zip object from the contents of a directory

        Given a directory name, generate an in-memory zip file
        containing all the files in the directory.  A bytes/string
        representation (suitable for use by from_zip_data()) can be
        obtained using the data() method, following a from_directory
        call.

        Parameters
        ----------
        pathname: str
            The directory of filenames to be placed in the context
            file.
        """
        self._buffer = io.BytesIO()
        the_file = zipfile.ZipFile(self._buffer, "w", compression=zipfile.ZIP_DEFLATED)
        for folder_name, _, file_names in os.walk(pathname):
            for filename in file_names:
                file_pathname = os.path.join(folder_name, filename)
                the_file.write(file_pathname, os.path.basename(file_pathname))
        self._set_type(the_file.namelist())
        the_file.close()

    @staticmethod
    def _fix_context_file(ctx_file: str) -> None:
        """Clean up a context file

        Currently, there is a bug in the single case context file saving code
        that puts information that cannot be recalled independently of other
        cases in the .ctx file.  Remove that information and rewrite the file.

        Parameters
        ----------
        ctx_file: str
            The name of the context file to process.
        """
        try:
            with open(ctx_file, "rb") as f:
                ctx = f.read()
            try:
                # Remove the Object MetaData block units, we do not want to restore them
                # as they could change.
                # These are the lines we need to remove:
                #  ensight.objs.core.VARIABLES.find('Pressure',
                #                                   attr="DESCRIPTION")[0].setmetatag('CFD_VAR','')
                #  ensight.objs.core.CURRENTCASE[0].setmetatag('ENS_UNITS_LABEL',2.000000)
                #  ensight.objs.core.CURRENTCASE[0].setmetatag('ENS_UNITS_DIMS',1.000000)
                #  ensight.objs.core.CURRENTCASE[0].setmetatag('ENS_UNITS_LABEL_TIME','s')
                #  ensight.objs.core.CURRENTCASE[0].setmetatag('ENS_UNITS_SYSTEM',1.000000)
                #  ensight.objs.core.CURRENTCASE[0].setmetatag('ENS_UNITS_SYSTEM_NAME','SI')
                start = ctx.index(b"# Object MetaData commands")
                end = ctx.index(b"# End Object MetaData commands")
                if (start >= 0) and (end >= 0):
                    saved_lines = list()
                    for line in ctx[start:end].split(b"\n"):
                        skip = b".setmetatag('CFD_VAR'" in line
                        skip = skip or (b".setmetatag('ENS_UNITS_LABEL'" in line)
                        skip = skip or (b".setmetatag('ENS_UNITS_DIMS'" in line)
                        skip = skip or (b".setmetatag('ENS_UNITS_LABEL_TIME'" in line)
                        skip = skip or (b".setmetatag('ENS_UNITS_SYSTEM'" in line)
                        skip = skip or (b".setmetatag('ENS_UNITS_SYSTEM_NAME'" in line)
                        skip = skip and line.startswith(b"ensight.objs.core.")
                        if skip:
                            continue
                        saved_lines.append(line)
                    ctx = ctx[:start] + b"\n".join(saved_lines) + ctx[end:]
            except ValueError:
                warnings.warn("Note: Object Metadata block not found")
            try:
                # remove the Textures block (as the textures are not in the project)
                start = ctx.index(b"# Textures")
                end = ctx.index(b"# Attributes To Restore Viewport Defaults")
                if (start >= 0) and (end >= 0):
                    ctx = ctx[:start] + ctx[end:]
            except ValueError:
                warnings.warn("Note: Object Metadata block not found")
            # rewrite the file
            with open(ctx_file, "wb") as f:
                f.write(ctx)
        except Exception as e:
            warnings.warn("Unable to filter out undesired context file content: {}".format(e))

    def _restore_context(self, ensight: Any) -> None:
        """Restore a context from the state in this object

        Unpack the zip contents to disk (temporary directory) and perform a context restore on
        the contents.

        Parameters
        ----------
        ensight : Any
            The EnSight interface to use to make the actual native API commands.

        """
        with tempfile.TemporaryDirectory() as tempdirname:
            the_file = zipfile.ZipFile(self._buffer, "r")
            the_file.extractall(path=tempdirname)
            if self._type in (self._SIMPLE_CONTEXT, self._FULL_CONTEXT):
                _ = ensight.file.context_restore_rescale("OFF")
                _ = ensight.file.restore_context(os.path.join(tempdirname, "context.ctx"))

    def _capture_context(
        self, ensight: Any, context: int = _SIMPLE_CONTEXT, all_cases: bool = True
    ) -> None:
        """Capture the current state

        Cause the EnSight interface to save a context into a temporary directory.
        Zip up the directory contents (along with a "type" marking file) into the
        zip object inside of this state instance.

        Parameters
        ----------
        ensight : Any
            The EnSight interface to use to make the actual native API commands.
        context : int, optional
            The type of context to save. By default, _SIMPLE_CONTEXT.
        all_cases : bool, optional
            By default, save all cases.  If all_cases is set to False, only
            the current case will be saved. By default, True.

        """
        with tempfile.TemporaryDirectory() as tempdirname:
            # Save a context
            which = "current_case"
            if all_cases:
                which = "all_cases"
            _ = ensight.file.save_context_type(which)
            _ = ensight.file.save_context(os.path.join(tempdirname, "context.ctx"))
            if context == self._SIMPLE_CONTEXT:
                # remove sections that cause problems
                with open(os.path.join(tempdirname, "simplecontext.txt"), "w") as fp:
                    fp.write("simplecontext")
                self._fix_context_file(os.path.join(tempdirname, "context.ctx"))
            elif context == self._FULL_CONTEXT:
                with open(os.path.join(tempdirname, "fullcontext.txt"), "w") as fp:
                    fp.write("fullcontext")
            self._build_from_directory(tempdirname)


def _capture_context(ensight: Any, full: bool) -> Any:
    """Private interface called by PyEnSight

    API that makes it simpler to capture a context from a PyEnSight session.

    Parameters
    ----------
    ensight: Any
        EnSight session interface
    full: bool
        True if a "full context" should be saved.
    Returns
    -------
    Any
        A base64 representation of the context.

    """
    tmp = EnsContext()
    mode = EnsContext._SIMPLE_CONTEXT
    if full:
        mode = EnsContext._FULL_CONTEXT
    tmp._capture_context(ensight, context=mode)
    return tmp._data(b64=True)


def _restore_context(ensight: Any, data: str) -> None:
    """Private interface called by PyEnSight

    API that makes it simpler to restore a context from a PyEnSight session.

    Parameters
    ----------
    ensight: Any
        EnSight session interface
    data: str
        A base64 representation of the context.
    """
    tmp = EnsContext()
    tmp._from_data(data)
    tmp._restore_context(ensight)



from typing import TYPE_CHECKING, Any, List, Optional, Union

if TYPE_CHECKING:
    try:
        import ensight
        from ensight.objs import ENS_QUERY  # type: ignore
    except ImportError:
        from ansys.api.pyensight import ensight_api
        from ansys.api.pyensight.ens_query import ENS_QUERY


class Query:
    """Provides the ``ensight.utils.query`` interface.

    The methods in this class implement simplified interfaces to common
    query-orientated operations, such as creating a temporal or temporal query.

    This class is instantiated as ``ensight.utils.query`` in EnSight Python
    and as ``Session.ensight.utils.query`` in PyEnSight. The constructor is
    passed the interface, which serves as the ``ensight`` module for either
    case. As a result, the methods can be accessed as ``ensight.utils.query.create_distance()``
    in EnSight Python or ``session.ensight.utils.query.create_distance()`` in PyEnSight.

    Parameters
    ----------
    interface :
        Entity that provides the ``ensight`` namespace. In the case of
        EnSight Python, the ``ensight`` module is passed. In the case
        of PyEnSight, ``Session.ensight`` is passed.
    """

    def __init__(self, interface: Union["ensight_api.ensight", "ensight"]):
        self._ensight = interface

    DISTANCE_PART1D: str = "1d_part"
    DISTANCE_LINE: str = "line_tool"
    DISTANCE_SPLINE: str = "spline"

    DIST_TYPE_LENGTH: str = "arc_length"
    DIST_TYPE_LENGTH_X: str = "x_arc_length"
    DIST_TYPE_LENGTH_Y: str = "y_arc_length"
    DIST_TYPE_LENGTH_Z: str = "z_arc_length"
    DIST_TYPE_ORIGIN: str = "from_origin"
    DIST_TYPE_ORIGIN_X: str = "x_from_origin"
    DIST_TYPE_ORIGIN_Y: str = "y_from_origin"
    DIST_TYPE_ORIGIN_Z: str = "z_from_origin"

    DIST_SEG_ACCUMULATION: str = "accumulation"
    DIST_SEG_RESET_EACH: str = "reset_each"

    def create_distance(
        self,
        name: str,
        query_type: str,
        part_list: List[Any],
        variable1: Any,
        variable2: Optional[Any] = None,
        num_samples: int = 20,
        point1: Optional[List[float]] = None,
        point2: Optional[List[float]] = None,
        distance_type: str = DIST_TYPE_LENGTH,
        segments_by: str = DIST_SEG_ACCUMULATION,
        spline_name: Optional[str] = None,
        new_plotter: bool = False,
    ) -> "ENS_QUERY":
        """Create an EnSight query over a distance object.

        Create an EnSight distance query using the passed parameters. Return
        the newly generated ``ENS_QUERY`` object.

        Parameters
        ----------
        name : str
            Name for the query.
        query_type : str
            Type of the query. This table describes the options:

            ================== ========================================
            Name               Query type
            ================== ========================================
            DISTANCE_PART1D    Samples along the points of a 1D part
            DISTANCE_LINE      Samples along the length of a line
            DISTANCE_SPLINE    Samples along the length of a spline
            ================== ========================================


        part_list: list
            List of parts to use as the source for the query. Part numbers,
            part names, and part objects are supported.
        variable1 :
            Variable to sample as the "y" value.
        variable2 : optional
            Variable to sample as the "x" value. The default is ``None``,
            in which case the "x" value is the distance along the sampling
            domain.
        num_samples : int, optional
            For a spline or line query, the number of samples to use along
            the length of the spline or line. The default is ``20``.
        point1 : list[float], optional
            For a line query, the ``x,y,z`` location of the start of the line.
            The default is ``None``.
        point2: : list[float], optional
            For a line query, the x,y,z location of the end of the line.
            The default is ``None``.
        distance_type : str, optional
            For a query over a distance (no second variable), how to compute
            distance. The default is ``"arc_length"``, in which case
            ``DIST-TYPE_LENGTH_X`` is used. This table describes
            the options:

            =================== =========================================
            Name                Query type
            =================== =========================================
            DIST_TYPE_LENGTH    Length along path
            DIST_TYPE_LENGTH_X  X component along path
            DIST_TYPE_LENGTH_Y  Y component along path
            DIST_TYPE_LENGTH_Z  Z component along path
            DIST_TYPE_ORIGIN    Distance from the origin
            DIST_TYPE_ORIGIN_X  X component of distance from the origin
            DIST_TYPE_ORIGIN_Y  Y component of distance from the origin
            DIST_TYPE_ORIGIN_Z  Z component of distance from the origin
            =================== =========================================


        segments_by : str, optional
            For a 1D part query, how to compute distance for the
            segments of the 1D part. The default is ``"accumulation"``,
            in which case ``DIST_SEG_ACCUMULATION`` is used.
            This table describes the options:

            ====================== ==========================================
            Name                   Segment handling
            ====================== ==========================================
            DIST_SEG_ACCUMULATION  Accumulate distance over segments
            DIST_SEG_RESET_EACH    Reset the distance value for each segment
            ====================== ==========================================


        spline_name : str, optional
            For a spline query, the name of the spline to sample along. The
            default is ``None``.
        new_plotter : bool, optional
            Whether to add the query to a newly created plotter. The default is
            ``False``.

        Returns
        -------
        ENS_QUERY
            ``ENS_QUERY`` instance on success or raises an exception on error.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> s.load_data(f"{s.cei_home}/ensight{s.cei_suffix}/data/cube/cube.case")
        >>> parts = s.ensight.objs.core.PARTS["Computational mesh"]
        >>> pnt1 = [-1.0, 0.5, 1.0]
        >>> pnt2 = [2.0, 0.5, 1.0]
        >>> query = s.ensight.utils.query.create_distance("Temperature vs Distance",
        >>>             s.ensight.utils.query.DISTANCE_LINE, parts, "temperature",
        >>>             point1=pnt1, point2=pnt2, new_plotter=True)

        """
        if query_type not in [self.DISTANCE_PART1D, self.DISTANCE_LINE, self.DISTANCE_SPLINE]:
            raise RuntimeError(f"Invalid query type: {query_type} specified.")

        var1 = self._get_variable(variable1)
        var2 = self._get_variable(variable2, "DISTANCE")

        if query_type == self.DISTANCE_LINE:
            if (point1 is None) or (point2 is None):
                raise RuntimeError("Both point1 and point2 must be specified.")
            self._create_query_core_begin(name, part_list)
            self._ensight.query_ent_var.number_of_sample_pts(num_samples)
            self._ensight.query_ent_var.constrain("line_tool")
            self._ensight.query_ent_var.line_loc(1, *point1)
            self._ensight.query_ent_var.line_loc(2, *point2)

        elif query_type == self.DISTANCE_PART1D:
            self._create_query_core_begin(name, part_list, single=True)
            self._ensight.query_ent_var.constrain("1d_part")
            self._ensight.query_ent_var.multiple_segments_by(segments_by)

        elif query_type == self.DISTANCE_SPLINE:
            if spline_name is None:
                raise RuntimeError("A spline_name must be specified.")
            self._create_query_core_begin(name, part_list)
            self._ensight.query_ent_var.number_of_sample_pts(num_samples)
            self._ensight.query_ent_var.constrain("spline")
            self._ensight.query_ent_var.spline_name(spline_name)

        self._ensight.query_ent_var.distance(distance_type)
        self._ensight.query_ent_var.variable_1(var1)
        self._ensight.query_ent_var.generate_over("distance")
        self._ensight.query_ent_var.variable_2(var2)

        query = self._create_query_core_end()
        if new_plotter:
            plot = self._ensight.objs.core.DEFAULTPLOT[0].createplotter()
            query.addtoplot(plot)
        return query

    TEMPORAL_NODE: str = "node"
    TEMPORAL_ELEMENT: str = "element"
    TEMPORAL_IJK: str = "ijk"
    TEMPORAL_XYZ: str = "cursor"
    TEMPORAL_MINIMUM: str = "min"
    TEMPORAL_MAXIMUM: str = "max"

    def create_temporal(
        self,
        name: str,
        query_type: str,
        part_list: List[Any],
        variable1: Any,
        variable2: Optional[Any] = None,
        start_time: Optional[float] = None,
        end_time: Optional[float] = None,
        num_samples: Optional[int] = None,
        node_id: Optional[int] = None,
        element_id: Optional[int] = None,
        xyz: Optional[List[float]] = None,
        ijk: Optional[List[int]] = None,
        new_plotter: bool = False,
    ) -> "ENS_QUERY":
        """Create an EnSight query over time.

        Parameters
        ----------
        name : str
            Name for the query.
        query_type:
            Type of the query. Descriptions of options follow.

            ================= =========================================================
            Name              Query type
            ================= =========================================================
            TEMPORAL_NODE     At a node number over time
            TEMPORAL_ELEMENT  At an element number over time
            TEMPORAL_IJK      At an ijk element over time
            TEMPORAL_XYZ      At some x,y,z location over time
            TEMPORAL_MINIMUM  Minimum value on the selected parts at each point in time
            TEMPORAL_MAXIMUM  Maximum value on the selected parts at each point in time
            ================= =========================================================


        part_list : list
            List of parts to use as the source for the query.
        variable1 :
            Variable to sample as the "y" value.
        variable2 : optional
            Variable to sample as the "x" value. The default is ``None``, in which
            case the "x" value is the distance along the sampling domain.
        start_time : float, optional
            Simulation time to start sampling at. The default is ``None``, in which
            case the first timestep, ``ENS_GLOBALS.SOLUTIONTIME_LIMITS[0]``, is used.
        end_time : float, optional
            The simulation time to end sampling at. The default is ``None``, in which
            case the last timestep, ``ENS_GLOBALS.SOLUTIONTIME_LIMITS[1]``, is used.
        num_samples : int, optional
            Number of samples to take along the selected timestep range. The default
            is ``None``, in which case the number of defined timesteps,
            ``ENS_GLOBALS.TIMESTEP_LIMITS[1] - ENS_GLOBALS.TIMESTEP_LIMITS[0] + 1``,
            is used.
        node_id : int, optional
            For a ``TEMPORAL_NODE`` query, the ID of the node to query. The default is
            ``None``.
        element_id : int, optional
            For a ``TEMPORAL_ELEMENT`` query, the ID of the element to query. The
            default is ``None``.
        xyz : list[float], optional
            For a ``TEMPORRAL_XYZ`` query, the ``[x,y,z]`` location in data space
            to query. The default is ``None``.
        ijk : list[int], optional
            For a ``TEMPORAL_IJK`` query, the ``[i,j,k]`` value for the structured
            mesh node to query.  The default is ``None``.
        new_plotter : bool, optional
            Whether to add the query to a newly created plotter. The default is
            ``False``.

        Returns
        -------
        ENS_QUERY
            ````ENS_QUERY`` instance on success or raises an exception on error.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> s.load_data(f"{s.cei_home}/ensight{s.cei_suffix}/data/guard_rail/crash.case")
        >>> parts = s.ensight.objs.core.PARTS
        >>> s.ensight.utils.query.create_temporal("Maximum plastic",
        >>> s.ensight.utils.query.TEMPORAL_MAXIMUM, parts, "plastic", new_plotter=True)

        """
        if query_type not in [
            self.TEMPORAL_NODE,
            self.TEMPORAL_ELEMENT,
            self.TEMPORAL_IJK,
            self.TEMPORAL_XYZ,
            self.TEMPORAL_MINIMUM,
            self.TEMPORAL_MAXIMUM,
        ]:
            raise RuntimeError(f"Invalid query type: {query_type} specified.")

        var1 = self._get_variable(variable1)
        var2 = self._get_variable(variable2, "TIME")

        # default the time range
        if start_time is None:
            start_time = self._ensight.objs.core.SOLUTIONTIME_LIMITS[0]
        if end_time is None:
            end_time = self._ensight.objs.core.SOLUTIONTIME_LIMITS[1]

        # default the number of timesteps
        if num_samples is None:
            num_samples = (
                self._ensight.objs.core.TIMESTEP_LIMITS[1]
                - self._ensight.objs.core.TIMESTEP_LIMITS[0]
                + 1
            )

        if query_type == self.TEMPORAL_NODE:
            if node_id is None:
                raise RuntimeError("node_id must be specified.")
            self._create_query_core_begin(name, part_list)
            self._ensight.query_ent_var.constrain("node")
            self._ensight.query_ent_var.node_id(node_id)

        elif query_type == self.TEMPORAL_ELEMENT:
            if element_id is None:
                raise RuntimeError("element_id must be specified.")
            self._create_query_core_begin(name, part_list)
            self._ensight.query_ent_var.constrain("element")
            self._ensight.query_ent_var.elem_id(element_id)

        elif query_type == self.TEMPORAL_XYZ:
            if xyz is None:
                raise RuntimeError("xyz must be specified.")
            self._create_query_core_begin(name, part_list)
            self._ensight.query_ent_var.constrain("cursor")
            self._ensight.query_ent_var.cursor_loc(*xyz)

        elif query_type == self.TEMPORAL_IJK:
            if ijk is None:
                raise RuntimeError("ijk must be specified.")
            self._create_query_core_begin(name, part_list)
            self._ensight.query_ent_var.constrain("ijk")
            self._ensight.query_ent_var.ijk(*ijk)

        elif query_type == self.TEMPORAL_MINIMUM:
            self._create_query_core_begin(name, part_list)
            self._ensight.query_ent_var.constrain("min")

        elif query_type == self.TEMPORAL_MAXIMUM:
            self._create_query_core_begin(name, part_list)
            self._ensight.query_ent_var.constrain("max")

        self._ensight.query_ent_var.number_of_sample_pts(num_samples)
        self._ensight.query_ent_var.begin_simtime(start_time)
        self._ensight.query_ent_var.end_simtime(end_time)
        self._ensight.query_ent_var.sample_by("value")
        self._ensight.query_ent_var.update_with_newtimesteps("ON")
        self._ensight.query_ent_var.variable_1(var1)
        self._ensight.query_ent_var.generate_over("time")
        self._ensight.query_ent_var.variable_2(var2)

        query = self._create_query_core_end()
        if new_plotter:
            plot = self._ensight.objs.core.DEFAULTPLOT[0].createplotter()
            query.addtoplot(plot)
        return query

    def _create_query_core_end(self) -> "ENS_QUERY":
        """Complete a query operation.

        Execute the "end()" and "query()" calls to finalize the query.  Return the latest
        ENS_QUERY object.

        Returns
        -------
        "ENS_QUERY"
            The last created ENS_QUERY object or raise an exception if query creation fails.

        """
        nextid = self._ensight.objs.next_id()
        self._ensight.query_ent_var.end()
        self._ensight.query_ent_var.query()
        error_msg = None
        try:
            query = max(self._ensight.objs.core.QUERIES)
            # no new id allocated
            if query.__OBJID__ < nextid:
                error_msg = "Unable to create the specified query."
        except ValueError:
            error_msg = "Unable to create the specified query."
        if error_msg:
            raise RuntimeError(error_msg)
        return query

    def _create_query_core_begin(self, name: str, parts: Optional[List[int]], single=False) -> None:
        """Common query setup

        Make the common calls for all queries.  Select the appropriate source parts, set the
        query name and "generated" type.

        Parameters
        ----------
        name : str
            The name for the query.
        parts : List
            The list of source parts for the query.
        single : bool
            If true, then this is a 1D part query and emit the appropriate output.  Otherwise,
            ensure that the proper parts are selected. (Default value = False)

        """
        part_list = []
        if parts:
            for p in parts:
                if type(p) == str:
                    part_list.append(self._ensight.objs.core.PARTS[p][0].PARTNUMBER)
                elif type(p) == int:
                    part_list.append(p)
                else:
                    if hasattr(p, "PARTNUMBER"):
                        part_list.append(p.PARTNUMBER)
        if not single:
            self._ensight.part.select_begin(part_list)
        self._ensight.query_ent_var.begin()
        self._ensight.query_ent_var.description(name)
        self._ensight.query_ent_var.query_type("generated")
        if single:
            self._ensight.query_ent_var.part_id(part_list[0])

    def _get_variable(self, v: Any, default: Optional[str] = None) -> Optional[str]:
        """Convert a generic argument into a variable name string

        Convert a generic object into a variable name. Valid inputs
        include a string or an ENS_VARIABLE object.

        Parameters
        ----------
        v: Any
            The object to be converted into a variable name.
        default: str, optional
            The value to return if the input variable is None.

        Returns
        -------
            The string name of the variable or None.

        """
        if v is None:
            return default
        elif type(v) is str:
            if v not in self._ensight.objs.core.VARIABLES.get_attr("DESCRIPTION"):
                raise ValueError("The variable supplied does not exist.")
            return v
        else:
            return v.DESCRIPTION



"""ensobjlist module

Emulation of the EnSight ensobjlist class

"""
from collections.abc import Iterable
import fnmatch
from typing import Any, List, Optional, TypeVar, no_type_check, overload

from ansys.pyensight.core.ensobj import ENSOBJ
from typing_extensions import SupportsIndex

T = TypeVar("T")


class ensobjlist(List[T]):  # noqa: N801
    """Class used when returning lists of EnSight proxy objects.  A subclass of 'list'.

    In the EnSight object Python bindings, whenever a list is returned that
    is known to exclusively contain ENSOBJ subclass objects, the ensobjlist
    (list subclass) is returned instead.  This class simply adds a few
    ENSOBJ specific methods and some functionality to the list object.

    One additional feature of the ensobjlist is that the __getitem__()
    interface supports strings and lists of strings.  In that situation,
    the object acts as if the find() method is called.

    These are equivalent

    >>> a = objlist["Hello"]
    >>> a = objlist.find("Hello", attr="DESCRIPTION")

    These are as well

    >>> a = objlist[("Hello", "Goodbye")]
    >>> a = objlist.find(("Hello", "Goodbye"), attr="DESCRIPTION")

    Parameters
    ----------
    \*args :
        Superclass (list) arguments
    \**kwargs :
        Superclass (list) keyword arguments


    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    @staticmethod
    def _is_iterable(arg: Any) -> bool:
        """Check if the arg is iterable, but not a string.

        Parameters
        ----------
        arg: Any
            Argument to check.

        """
        return isinstance(arg, Iterable) and not isinstance(arg, str)

    def find(
        self, value: Any, attr: Any = "DESCRIPTION", group: int = 0, wildcard: int = 0
    ) -> "ensobjlist[T]":
        """Find objects in the list using the ENSOBJ interface.

        This method will scan the ENSOBJ subclass objects in the list and return
        an ensobjlist of those matching the search criteria.

        Parameters
        ----------
        value: Any
            A single object or a tuple of objects that will be compared to the value of
            an attribute accessed via the getattr() ENSOBJ interface.
        attr: Any
            The specific attribute (id or string) to look up using getattr().
        group: int
            Currently unimplemented.
        wildcard: int
            Instead of the comparison being done via the equals test, it will be done using
            fnmatch between the string representation of the item and the value.  This allows
            values to be specified using glob wildcard specifications.  If wildcard is set
            to 1, this is a case-sensitive wildcard operation.  If set to 2, the comparison
            is case-insensitive.  The default is not to use wildcard comparisons (0).

        Returns
        -------
        ensobjlist[T]
            An ensobjlist of the items that matched the search criteria.

        """
        value_list = value
        if not self._is_iterable(value):
            value_list = [value]
        out_list: ensobjlist[Any] = ensobjlist()
        for item in self:
            if isinstance(item, ENSOBJ):
                try:
                    item_value = item.getattr(attr)
                    for check_value in value_list:
                        if wildcard == 2:
                            if fnmatch.fnmatch(str(item_value), str(check_value)):
                                out_list.append(item)
                                break
                        elif wildcard > 0:
                            if fnmatch.fnmatchcase(str(item_value), str(check_value)):
                                out_list.append(item)
                                break
                        else:
                            if item_value == check_value:
                                out_list.append(item)
                                break
                except RuntimeError:
                    pass
        # TODO: handle group
        return out_list

    def set_attr(self, attr: Any, value: Any) -> int:
        """Set an attribute value on all contained objects

        Walk the items in this object.  If they are ENSOBJ subclasses, attempt to set
        the specified attribute id to the specified value.  Count the number of times
        that operation was successful and return that number.

        Parameters
        ----------
        attr: Any
            The specific attribute (id or string) to change using setattr().
        value: Any
            The value to attempt to set the specified attribute to.

        Returns
        -------
        int
            The number of successful set operations.

        Examples
        --------
        >>> session.ensight.objs.core.PARTS.set_attr("VISIBLE", True)

        """
        objid_list = []
        session = None
        objid_list = [x.__OBJID__ for x in self if isinstance(x, ENSOBJ)]
        for item in self:
            if hasattr(item, "_session"):
                session = item._session
                break
        if session:
            msg = f"ensight.objs.ensobjlist(ensight.objs.wrap_id(x) for x in {objid_list}).set_attr({attr.__repr__()}, {value.__repr__()})"
            return session.cmd(msg)
        return 0

    def get_attr(self, attr: Any, default: Optional[Any] = None):
        """Query a specific attribute for all ENSOBJ objects in the list

        Walk the items in this object.  If they are ENSOBJ subclasses, query the value of
        the passed attribute id.  If the item is not an ENSOBJ subclass or the attribute
        query fails, the returned list will have the specified default value for that item.

        Parameters
        ----------
        attr: Any
            The specific attribute (id or string) to look up using getattr().
        default: Any, optional
            The value to return for objects that are not ENSOBJ subclasses or do not
            support the specified attribute.

        Returns
        -------
        List
            A list of the attribute values for each item in this object

        Examples
        --------
        >>> state = session.ensight.core.PARTS.get_attr(session.ensight.objs.enums.VISIBLE)

        """
        objid_list = []
        session = None
        objid_list = [x.__OBJID__ for x in self if isinstance(x, ENSOBJ)]
        for item in self:
            if hasattr(item, "_session"):
                session = item._session
                break
        value = None
        if session:
            if default:
                msg = f"ensight.objs.ensobjlist(ensight.objs.wrap_id(x) for x in {objid_list}).get_attr({attr.__repr__()}, {default.__repr__()})"
            else:
                msg = f"ensight.objs.ensobjlist(ensight.objs.wrap_id(x) for x in {objid_list}).get_attr({attr.__repr__()})"
            value = session.cmd(msg)
        if value:
            return value
        return [default] * len(objid_list)

    @overload
    def __getitem__(self, index: SupportsIndex) -> T:
        ...

    @overload
    def __getitem__(self, index: slice) -> List[T]:
        ...

    def __getitem__(self, index):
        """Overload the getitem operator to allow for tuple and string DESCRIPTION queries"""
        if isinstance(index, str) or isinstance(index, tuple):
            return self.find(index)
        return super().__getitem__(index)

    def __str__(self):
        ret_str = ", ".join([str(x) for x in self])
        return f"[{ret_str}]"

    @no_type_check
    def _repr_pretty_(self, p: "pretty", cycle: bool) -> None:
        """Support the pretty module for better IPython support

        Parameters
        ----------
        p: "pretty" :
            pretty flag.
        cycle: bool :
            cycle flag.

        """
        name = self.__class__.__name__
        if cycle:
            p.text(f"{name}(...)")
        else:
            with p.group(len(name) + 2, f"{name}([", "])"):
                for idx, item in enumerate(self):
                    if idx:
                        p.text(",")
                        p.breakable()
                    p.pretty(item)



try:
    import importlib.metadata as importlib_metadata  # type: ignore
except ModuleNotFoundError:  # pragma: no cover
    import importlib_metadata  # type: ignore
__version__ = importlib_metadata.version(__name__.replace(".", "-"))

VERSION = __version__
DEFAULT_ANSYS_VERSION = "241"

# Ansys version number that this release is associated with
__ansys_version__ = DEFAULT_ANSYS_VERSION
__ansys_version_str__ = f"{2000+(int(__ansys_version__) // 10)} R{int(__ansys_version__) % 10}"

from ansys.pyensight.core.dockerlauncher import DockerLauncher
from ansys.pyensight.core.launch_ensight import launch_ensight
from ansys.pyensight.core.launcher import Launcher
from ansys.pyensight.core.listobj import ensobjlist
from ansys.pyensight.core.locallauncher import LocalLauncher
from ansys.pyensight.core.session import Session



"""ensight_grpc module

This package defines the EnSightGRPC class which provides a simpler
interface to the EnSight gRPC interface, including event streams.

"""
import threading
from typing import Any, Callable, List, Optional, Tuple, Union
import uuid

from ansys.api.pyensight.v0 import ensight_pb2, ensight_pb2_grpc
import grpc


class EnSightGRPC(object):
    """Wrapper around a gRPC connection to an EnSight instance

    This class provides an asynchronous interface to the EnSight
    core gRPC interface.  It can handle remote event
    streams, providing a much simpler interface to the EnSight
    application. The default is to make a connection to an EnSight
    gRPC server on port 12345 on the loopback host.

    Parameters
    ----------
    host: str, optional
        Hostname where there EnSight gRPC server is running.
    port: int, optional
        Port to make the gRPC connection to
    secret_key: str, optional
        Connection secret key
    """

    def __init__(self, host: str = "127.0.0.1", port: int = 12345, secret_key: str = ""):
        self._host = host
        self._port = port
        self._channel = None
        self._stub = None
        self._dsg_stub = None
        self._security_token = secret_key
        # Streaming APIs
        # Event (strings)
        self._event_stream = None
        self._event_thread: Optional[threading.Thread] = None
        self._events: List[Any] = list()
        # Callback for events (self._events not used)
        self._event_callback: Optional[Callable] = None
        self._prefix: Optional[str] = None

    @property
    def host(self) -> str:
        """The gRPC server (EnSight) hostname"""
        return self._host

    def port(self) -> int:
        """The gRPC server (EnSight) port number"""
        return self._port

    @property
    def security_token(self) -> str:
        """The gRPC server (EnSight) secret key

        EnSight supports a security token in either numeric (-security {int}) or
        string (ENSIGHT_SECURITY_TOKEN environmental variable) form.  If EnSight
        is using a security token, all gRPC calls must include this token.  This
        call sets the token for all grPC calls made by this class.
        """
        return self._security_token

    @security_token.setter
    def security_token(self, name: str) -> None:
        self._security_token = name

    def shutdown(self, stop_ensight: bool = False, force: bool = False) -> None:
        """Close down the gRPC connection

        Disconnect all connections to the gRPC server.  If stop_ensight is True, send the
        'Exit' command to the EnSight gRPC server.

        Parameters
        ----------
        stop_ensight: bool, optional
            if True, send an 'Exit' command to the gRPC server.
        force: bool, optional
            if stop_ensight and force are true, stop EnSight aggressively
        """
        if self.is_connected():
            # if requested, send 'Exit'
            if stop_ensight:
                # the gRPC ExitRequest is exactly that, a request in some
                # cases the operation needs to be forced
                if force:
                    try:
                        self.command("ensight.exit(0)", do_eval=False)
                    except IOError:
                        # we expect this as the exit can result in the gRPC call failing
                        pass
                else:
                    if self._stub:
                        _ = self._stub.Exit(ensight_pb2.ExitRequest(), metadata=self._metadata())
            # clean up control objects
            self._stub = None
            self._dsg_stub = None
            if self._channel:
                self._channel.close()
            self._channel = None

    def is_connected(self) -> bool:
        """Check to see if the gRPC connection is live

        Returns
        -------
             True if the connection is active.
        """
        return self._channel is not None

    def connect(self, timeout: float = 15.0) -> None:
        """Establish the gRPC connection to EnSight

        Attempt to connect to an EnSight gRPC server using the host and port
        established by the constructor.  Note on failure, this function just
        returns, but is_connected() will return False.

        Parameters
        ----------
        timeout: float
            how long to wait for the connection to timeout
        """
        if self.is_connected():
            return
        # set up the channel
        self._channel = grpc.insecure_channel(
            "{}:{}".format(self._host, self._port),
            options=[
                ("grpc.max_receive_message_length", -1),
                ("grpc.max_send_message_length", -1),
                ("grpc.testing.fixed_reconnect_backoff_ms", 1100),
            ],
        )
        try:
            grpc.channel_ready_future(self._channel).result(timeout=timeout)
        except grpc.FutureTimeoutError:
            self._channel = None
            return
        # hook up the stub interface
        self._stub = ensight_pb2_grpc.EnSightServiceStub(self._channel)

    def _metadata(self) -> List[Tuple[bytes, Union[str, bytes]]]:
        """Compute the gRPC stream metadata

        Compute the list to be passed to the gRPC calls for things like security.
        """
        ret: List[Tuple[bytes, Union[str, bytes]]] = list()
        s: Union[str, bytes]
        if self._security_token:
            s = self._security_token
            if type(s) == str:
                s = s.encode("utf-8")
            ret.append((b"shared_secret", s))
        return ret

    def render(
        self,
        width: int = 640,
        height: int = 480,
        aa: int = 1,
        png: bool = True,
        highlighting: bool = False,
    ) -> bytes:
        """Generate a rendering of the current EnSight scene

        Render the current scene at a specific size and using a specific number of anti-aliasing
        passes. The return value can be a byte array (width*height*3) bytes or a PNG image.

        Parameters
        ----------
        width: int, optional
            width of the image to render
        height: int, optional
            height of the image to render
        aa: int, optional
            number of antialiasing passes to use in generating the image
        png: bool, optional
            if True, the return value is a PNG image bytestream.  Otherwise, it is a simple
            bytes object with width*height*3 values.
        highlighting: bool, optional
            if True, selection highlighting will be included in the image.

        Returns
        -------
        bytes
            bytes object representation of the rendered image

        Raises
        ------
            IOError if the operation fails
        """
        self.connect()
        ret_type = ensight_pb2.RenderRequest.IMAGE_RAW
        if png:
            ret_type = ensight_pb2.RenderRequest.IMAGE_PNG
        response: Any
        try:
            if self._stub:
                response = self._stub.RenderImage(
                    ensight_pb2.RenderRequest(
                        type=ret_type,
                        image_width=width,
                        image_height=height,
                        image_aa_passes=aa,
                        include_highlighting=highlighting,
                    ),
                    metadata=self._metadata(),
                )
        except Exception:
            raise IOError("gRPC connection dropped")
        return response.value

    def geometry(self) -> bytes:
        """Return the current scene geometry in glTF format

        Package up the geometry currently being viewed in the EnSight session as
        a glTF stream.  Return this stream as an array of byte.  Note: no
        intermediate files are utilized.

        Note: currently there is a limitation of glTF files to 2GB

        Returns
        -------
            bytes object representation of the glTF file

        Raises
        ------
            IOError if the operation fails
        """
        self.connect()
        response: Any
        try:
            if self._stub:
                response = self._stub.GetGeometry(
                    ensight_pb2.GeometryRequest(type=ensight_pb2.GeometryRequest.GEOMETRY_GLB),
                    metadata=self._metadata(),
                )
        except Exception:
            raise IOError("gRPC connection dropped")
        return response.value

    def command(self, command_string: str, do_eval: bool = True, json: bool = False) -> Any:
        """Send a Python command string to be executed in EnSight

        The string will be run or evaluated in the EnSight Python interpreter via the
        EnSightService::RunPython() gRPC all.  If an exception or other error occurs, this
        function will throw a RuntimeError.  If do_eval is False, the return value will be None,
        otherwise it will be the returned string (eval() will not be performed).  If json is True,
        the return value will be a JSON representation of the report execution result.

        Parameters
        ----------
        command_string: str
            The string to execute
        do_eval: bool, optional
            If True, a return value will be computed and returned
        json: bool, optional
            If True and do_eval is True, the return value will be a JSON representation of
            the evaluated value.

        Returns
        -------
        Any
             None, a string ready for Python eval() or a JSON string.

        Raises
        ------
            RuntimeError if the operation fails.
            IOError if the communication fails.
        """
        self.connect()
        flags = ensight_pb2.PythonRequest.EXEC_RETURN_PYTHON
        response: Any
        if json:
            flags = ensight_pb2.PythonRequest.EXEC_RETURN_JSON
        if not do_eval:
            flags = ensight_pb2.PythonRequest.EXEC_NO_RESULT
        try:
            if self._stub:
                response = self._stub.RunPython(
                    ensight_pb2.PythonRequest(type=flags, command=command_string),
                    metadata=self._metadata(),
                )
        except Exception:
            raise IOError("gRPC connection dropped")
        if response.error < 0:
            raise RuntimeError(response.value)
        if flags == ensight_pb2.PythonRequest.EXEC_NO_RESULT:
            return None
        # This was moved externally so pre-processing could be performed
        # elif flags == ensight_pb2.PythonRequest.EXEC_RETURN_PYTHON:
        #    return eval(response.value)
        return response.value

    def prefix(self) -> str:
        """Return the unique prefix for this instance.

        Some EnSight gRPC APIs require a unique prefix so that EnSight can handle
        multiple, simultaneous remote connections. This method will generate a GUID-based
        prefix.

        Returns
        -------
        str
            A unique (for this session) prefix string of the form: grpc://{uuid}/
        """
        # prefix URIs will have the format:  "grpc://{uuid}/{callbackname}?enum={}&uid={}"
        if self._prefix is None:
            self._prefix = "grpc://" + str(uuid.uuid1()) + "/"
        return self._prefix

    def event_stream_enable(self, callback: Optional[Callable] = None) -> None:
        """Enable a simple gRPC-based event stream from EnSight

        This method makes a EnSightService::GetEventStream() gRPC call into EnSight, returning
        an ensightservice::EventReply stream.  The method creates a thread to hold this
        stream open and read new events from it.  The thread adds the event strings to
        a list of events stored on this instance.  If callback is not None, the object
        will be called with the event string, otherwise they can be retrieved using get_event().
        """
        if self._event_stream is not None:
            return
        self._event_callback = callback
        self.connect()
        if self._stub:
            self._event_stream = self._stub.GetEventStream(
                ensight_pb2.EventStreamRequest(prefix=self.prefix()),
                metadata=self._metadata(),
            )
        self._event_thread = threading.Thread(target=self._poll_events)
        self._event_thread.daemon = True
        self._event_thread.start()

    def event_stream_is_enabled(self) -> bool:
        """Check to see if the event stream is enabled

        If an event stream has been successfully established via
        event_stream_enable(), then this function returns True.

        Returns
        -------
              True if a ensightservice::EventReply steam is active
        """
        return self._event_stream is not None

    def get_event(self) -> Optional[str]:
        """Retrieve and remove the oldest ensightservice::EventReply string

        When any of the event streaming systems is enabled, Python threads will receive the
        event records and store them in this instance in an ordered fashion.  This method
        retrieves the oldest ensightservice::EventReply string in the queue.

        Returns
        -------
            None or the oldest event string in the queue.
        """
        try:
            return self._events.pop(0)
        except IndexError:
            return None

    def _put_event(self, evt: "ensight_pb2.EventReply") -> None:
        """Add an event record to the event queue on this instance

        This method is used by threads to make the events they receive available to
        calling applications via get_event().
        """
        if self._event_callback:
            self._event_callback(evt.tag)
            return
        self._events.append(evt.tag)

    def _poll_events(self) -> None:
        """Internal method to handle event streams

        This method is called by a Python thread to read events via the established
        ensightservice::EventReply stream.
        """
        try:
            while self._stub is not None:
                evt = self._event_stream.next()
                self._put_event(evt)
        except Exception:
            # signal that the gRPC connection has broken
            self._event_stream = None
            self._event_thread = None



import os
import tempfile
from types import ModuleType
from typing import Any, Optional, Union
import uuid

from PIL import Image
import numpy

try:
    import ensight
    import enve
except ImportError:
    from ansys.api.pyensight import ensight_api


class Export:
    """Provides the ``ensight.utils.export`` interface.

    The methods in this class implement simplified interfaces to common
    image and animation export operations.

    This class is instantiated as ``ensight.utils.export`` in EnSight Python
    and as ``Session.ensight.utils.export`` in PyEnSight. The constructor is
    passed the interface, which serves as the ``ensight`` module for either
    case. As a result, the methods can be accessed as ``ensight.utils.export.image()``
    in EnSight Python or ``session.ensight.utils.export.animation()`` in PyEnSight.

    Parameters
    ----------
    interface :
        Entity that provides the ``ensight`` namespace. In the case of
        EnSight Python, the ``ensight`` module is passed. In the case
        of PyEnSight, ``Session.ensight`` is passed.
    """

    def __init__(self, interface: Union["ensight_api.ensight", "ensight"]):
        self._ensight = interface

    def _remote_support_check(self):
        """Determine if ``ensight.utils.export`` exists on the remote system.

            Before trying to use this module, use this method to determine if this
            module is available in the EnSight instance.

        Raises
        ------
            RuntimeError if the module is not present.
        """
        # if a module, then we are inside EnSight
        if isinstance(self._ensight, ModuleType):
            return
        try:
            _ = self._ensight._session.cmd("dir(ensight.utils.export)")
        except RuntimeError:
            import ansys.pyensight.core

            raise RuntimeError(
                f"Remote EnSight session must have PyEnsight version \
            {ansys.pyensight.core.DEFAULT_ANSYS_VERSION} or higher installed to use this API."
            )

    TIFFTAG_IMAGEDESCRIPTION: int = 0x010E

    def image(
        self,
        filename: str,
        width: Optional[int] = None,
        height: Optional[int] = None,
        passes: int = 4,
        enhanced: bool = False,
        raytrace: bool = False,
    ) -> None:
        """Render an image of the current EnSight scene.

        This method returns a PIL image object.

        Parameters
        ----------
        filename : str
            Name of the local file to save the image to.
        width : int, optional
            Width of the image in pixels. The default is ``None``, in which case
            ```ensight.objs.core.WINDOWSIZE[0]`` is used.
        height : int, optional
            Height of the image in pixels. The default is ``None``, in which case
            ``ensight.objs.core.WINDOWSIZE[1]`` is used.
        passes : int, optional
            Number of antialiasing passes. The default is ``4``.
        enhanced : bool, optional
            Whether to save the image to the filename specified in the TIFF format.
            The default is ``False``. The TIFF format includes additional channels
            for the per-pixel object and variable information.
        raytrace : bool, optional
            Whether to render the image with the raytracing engine. The default is ``False``.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> s.load_data(f"{s.cei_home}/ensight{s.cei_suffix}/data/cube/cube.case")
        >>> s.ensight.utils.export.image("example.png")

        """
        self._remote_support_check()

        win_size = self._ensight.objs.core.WINDOWSIZE
        if width is None:
            width = win_size[0]
        if height is None:
            height = win_size[1]

        if isinstance(self._ensight, ModuleType):
            raw_image = self._image_remote(width, height, passes, enhanced, raytrace)
        else:
            cmd = f"ensight.utils.export._image_remote({width}, {height}, {passes}, "
            cmd += f"{enhanced}, {raytrace})"
            raw_image = self._ensight._session.cmd(cmd)

        pil_image = self._dict_to_pil(raw_image)
        if enhanced:
            tiffinfo_dir = {self.TIFFTAG_IMAGEDESCRIPTION: raw_image["metadata"]}
            pil_image[0].save(
                filename,
                save_all=True,
                append_images=[pil_image[1], pil_image[2]],
                tiffinfo=tiffinfo_dir,
            )
        else:
            pil_image[0].save(filename)

    def _dict_to_pil(self, data: dict) -> list:
        """Convert the contents of the dictionary into a PIL image.

        Parameters
        ----------
        data : dict
            Dictionary representation of the contents of the ``enve`` object.

        Returns
        -------
        list
            List of one or three image objects, [RGB {, pick, variable}].
        """
        images = [
            Image.fromarray(self._numpy_from_dict(data["pixeldata"])).transpose(
                Image.FLIP_TOP_BOTTOM
            )
        ]
        if data.get("variabledata", None) and data.get("pickdata", None):
            images.append(
                Image.fromarray(self._numpy_from_dict(data["pickdata"])).transpose(
                    Image.FLIP_TOP_BOTTOM
                )
            )
            images.append(
                Image.fromarray(self._numpy_from_dict(data["variabledata"])).transpose(
                    Image.FLIP_TOP_BOTTOM
                )
            )
        return images

    @staticmethod
    def _numpy_to_dict(array: Any) -> Optional[dict]:
        """Convert a numpy array into a dictionary.

        Parameters
        ----------
        array:
            Numpy array or None.

        Returns
        -------
        ``None`` or a dictionary that can be serialized.
        """
        if array is None:
            return None
        return dict(shape=array.shape, dtype=array.dtype.str, data=array.tostring())

    @staticmethod
    def _numpy_from_dict(obj: Optional[dict]) -> Any:
        """Convert a dictionary into a numpy array.

        Parameters
        ----------
        obj:
            Dictionary generated by ``_numpy_to_dict`` or ``None``.

        Returns
        -------
        ``None`` or a numpy array.
        """
        if obj is None:
            return None
        return numpy.frombuffer(obj["data"], dtype=obj["dtype"]).reshape(obj["shape"])

    def _image_remote(
        self, width: int, height: int, passes: int, enhanced: bool, raytrace: bool
    ) -> dict:
        """EnSight-side implementation.

        Parameters
        ----------
        width : int
            Width of the image in pixels.
        height : int
            Height of the image in pixels.
        passes : int
            Number of antialiasing passes.
        enhanced : bool
            Whether to returned the image as a "deep pixel" TIFF image file.
        raytrace :
            Whether to render the image with the raytracing engine.

        Returns
        -------
        dict
            Dictionary of the various channels.
        """
        if not raytrace:
            img = ensight.render(x=width, y=height, num_samples=passes, enhanced=enhanced)
        else:
            with tempfile.TemporaryDirectory() as tmpdirname:
                tmpfilename = os.path.join(tmpdirname, str(uuid.uuid1()))
                ensight.file.image_format("png")
                ensight.file.image_file(tmpfilename)
                ensight.file.image_window_size("user_defined")
                ensight.file.image_window_xy(width, height)
                ensight.file.image_rend_offscreen("ON")
                ensight.file.image_numpasses(passes)
                ensight.file.image_stereo("current")
                ensight.file.image_screen_tiling(1, 1)
                ensight.file.raytracer_options("fgoverlay 1 imagedenoise 1 quality 5")
                ensight.file.image_raytrace_it("ON")
                ensight.file.save_image()
                img = enve.image()
                img.load(f"{tmpfilename}.png")
        # get the channels from the enve.image instance
        output = dict(width=width, height=height, metadata=img.metadata)
        # extract the channels from the image
        output["pixeldata"] = self._numpy_to_dict(img.pixeldata)
        output["variabledata"] = self._numpy_to_dict(img.variabledata)
        output["pickdata"] = self._numpy_to_dict(img.pickdata)
        return output

    ANIM_TYPE_SOLUTIONTIME: int = 0
    ANIM_TYPE_ANIMATEDTRACES: int = 1
    ANIM_TYPE_FLIPBOOK: int = 2
    ANIM_TYPE_KEYFRAME: int = 3

    def animation(
        self,
        filename: str,
        width: Optional[int] = None,
        height: Optional[int] = None,
        passes: int = 4,
        anim_type: int = ANIM_TYPE_SOLUTIONTIME,
        frames: Optional[int] = None,
        starting_frame: int = 0,
        frames_per_second: float = 60.0,
        format_options: Optional[str] = "",
        raytrace: bool = False,
    ) -> None:
        """Generate an MPEG4 animation file.

        An MPEG4 animation file can be generated from temporal data, flipbooks, keyframes,
        or animated traces.

        Parameters
        ----------
        filename : str
            Name for the MPEG4 file to save to local disk.
        width : int, optional
            Width of the image in pixels. The default is ``None``, in which case
            ``ensight.objs.core.WINDOWSIZE[0]`` is used.
        height : int, optional
            Height of the image in pixels. The default is ``None``, in which case
            ``ensight.objs.core.WINDOWSIZE[1]`` is used.
        passes : int, optional
            Number of antialiasing passes. The default  is ``4``.
        anim_type : int, optional
            Type of the animation to render. The default is ``0``, in which case
            ``"ANIM_TYPE_SOLUTIONTIME"`` is used. This table provides descriptions
            by each option number and name:

            ========================== ========================================
            Name                       Animation type
            ========================== ========================================
            0: ANIM_TYPE_SOLUTIONTIME   Animation over all solution times
            1: ANIM_TYPE_ANIMATEDTRACES Records animated rotations and traces
            2: ANIM_TYPE_FLIPBOOK       Records current flipbook animation
            3: ANIM_TYPE_KEYFRAME       Records current kKeyframe animation
            ========================== ========================================

        frames : int, optional
            Number of frames to save. The default is ``None``. The default for
            all but ``ANIM_TYPE_ANIMATEDTRACES`` covers all timesteps, flipbook
            pages, or keyframe steps. If ``ANIM_TYPE_ANIMATEDTRACES`` is specified,
            this keyword is required.
        starting_frame : int, optional
            Keyword for saving a subset of the complete collection of frames.
            The default is ``0``.
        frames_per_second : float, optional
            Number of frames per second for playback in the saved animation.
            The default is ``60.0``.
        format_options : str, optional
            More specific options for the MPEG4 encoder. The default is ``""``.
        raytrace : bool, optional
            Whether to render the image with the raytracing engine. The default is ``False``.

        Examples
        --------
        >>> s = LocalLauncher().start()
        >>> data = f"{s.cei_home}/ensight{s.cei_suffix}gui/demos/Crash Queries.ens"
        >>> s.ensight.objs.ensxml_restore_file(data)
        >>> quality = "Quality Best Type 1"
        >>> s.ensight.utils.export.animation("local_file.mp4", format_options=quality)

        """
        self._remote_support_check()

        win_size = self._ensight.objs.core.WINDOWSIZE
        if width is None:
            width = win_size[0]
        if height is None:
            height = win_size[1]

        if format_options is None:
            format_options = "Quality High Type 1"

        num_frames: int = 0
        if frames is None:
            if anim_type == self.ANIM_TYPE_SOLUTIONTIME:
                num_timesteps = self._ensight.objs.core.TIMESTEP_LIMITS[1]
                num_frames = num_timesteps - starting_frame
            elif anim_type == self.ANIM_TYPE_ANIMATEDTRACES:
                raise RuntimeError("frames is a required keyword with ANIMATEDTRACES animations")
            elif anim_type == self.ANIM_TYPE_FLIPBOOK:
                num_flip_pages = len(self._ensight.objs.core.FLIPBOOKS[0].PAGE_DETAILS)
                num_frames = num_flip_pages - starting_frame
            elif anim_type == self.ANIM_TYPE_KEYFRAME:
                num_keyframe_pages = self._ensight.objs.core.KEYFRAMEDATA["totalFrames"]
                num_frames = num_keyframe_pages - starting_frame
        else:
            num_frames = frames

        if num_frames < 1:
            raise RuntimeError(
                "No frames selected. Perhaps a static dataset SOLUTIONTIME request \
                 or no FLIPBOOK/KEYFRAME defined."
            )

        if isinstance(self._ensight, ModuleType):
            raw_mpeg4 = self._animation_remote(
                width,
                height,
                passes,
                anim_type,
                starting_frame,
                num_frames,
                frames_per_second,
                format_options,
                raytrace,
            )
        else:
            cmd = f"ensight.utils.export._animation_remote({width}, {height}, {passes}, "
            cmd += f"{anim_type}, {starting_frame}, {num_frames}, "
            cmd += f"{frames_per_second}, '{format_options}', {raytrace})"
            raw_mpeg4 = self._ensight._session.cmd(cmd)

        with open(filename, "wb") as fp:
            fp.write(raw_mpeg4)

    def _animation_remote(
        self,
        width: int,
        height: int,
        passes: int,
        anim_type: int,
        start: int,
        frames: int,
        fps: float,
        options: str,
        raytrace: bool,
    ) -> bytes:
        """EnSight-side implementation.

        Parameters
        ----------
        width : int
            Width of the image in pixels.
        height : int
            Height of the image in pixels.
        passes : int
            Number of antialiasing passes.
        anim_type : int
            Type of animation to save.
        start : int
            First frame number to save.
        frames : int
            Number of frames to save.
        fps : float
            Output framerate.
        options : str
            MPEG4 configuration options.
        raytrace : bool
            Whether to render the image with the raytracing engine.

        Returns
        -------
        bytes
            MPEG4 stream in bytes.
        """

        with tempfile.TemporaryDirectory() as tmpdirname:
            tmpfilename = os.path.join(tmpdirname, str(uuid.uuid1()) + ".mp4")
            self._ensight.file.animation_rend_offscreen("ON")
            self._ensight.file.animation_screen_tiling(1, 1)
            self._ensight.file.animation_format("mpeg4")
            if options:
                self._ensight.file.animation_format_options(options)
            self._ensight.file.animation_frame_rate(fps)
            self._ensight.file.animation_rend_offscreen("ON")
            self._ensight.file.animation_numpasses(passes)
            self._ensight.file.animation_stereo("mono")
            self._ensight.file.animation_screen_tiling(1, 1)
            self._ensight.file.animation_file(tmpfilename)
            self._ensight.file.animation_window_size("user_defined")
            self._ensight.file.animation_window_xy(width, height)
            self._ensight.file.animation_frames(frames)
            self._ensight.file.animation_start_number(start)
            self._ensight.file.animation_multiple_images("OFF")
            if raytrace:
                self._ensight.file.animation_raytrace_it("ON")
            else:
                self._ensight.file.animation_raytrace_it("OFF")
            self._ensight.file.animation_raytrace_ext("OFF")

            self._ensight.file.animation_play_time("OFF")
            self._ensight.file.animation_play_flipbook("OFF")
            self._ensight.file.animation_play_keyframe("OFF")

            self._ensight.file.animation_reset_time("OFF")
            self._ensight.file.animation_reset_traces("OFF")
            self._ensight.file.animation_reset_flipbook("OFF")
            self._ensight.file.animation_reset_keyframe("OFF")

            if anim_type == self.ANIM_TYPE_SOLUTIONTIME:
                # playing over time
                self._ensight.file.animation_play_time("ON")
                self._ensight.file.animation_reset_time("ON")
            elif anim_type == self.ANIM_TYPE_ANIMATEDTRACES:
                # recording particle traces/etc
                self._ensight.file.animation_reset_traces("ON")
            elif anim_type == self.ANIM_TYPE_KEYFRAME:
                self._ensight.file.animation_reset_keyframe("ON")
                self._ensight.file.animation_play_keyframe("ON")
            elif anim_type == self.ANIM_TYPE_FLIPBOOK:
                self._ensight.file.animation_play_flipbook("ON")
                self._ensight.file.animation_reset_flipbook("ON")

            self._ensight.file.save_animation()

            with open(tmpfilename, "rb") as fp:
                mp4_data = fp.read()

        return mp4_data



from typing import TYPE_CHECKING, Any, ContextManager, Union

if TYPE_CHECKING:
    try:
        import ensight
    except ImportError:
        from ansys.api.pyensight import ensight_api


class Support:
    """Provides the ``ensight.utils.support`` interface.

    This class provides a collection of general utility functions and objects
    that can be used to simplify various issues that come up when using EnSight
    and PyEnSight.

    This class is instantiated as ``ensight.utils.support`` in EnSight Python
    and as ``Session.ensight.utils.support`` in PyEnSight. The constructor is
    passed the interface, which serves as the ``ensight`` module for either
    case. As a result, the methods can be accessed as ``ensight.utils.support.*``
    in EnSight Python or ``session.ensight.utils.support.*`` in PyEnSight.

    Parameters
    ----------
    interface :
        Entity that provides the ``ensight`` namespace. In the case of
        EnSight Python, the ``ensight`` module is passed. In the case
        of PyEnSight, ``Session.ensight`` is passed.
    """

    def __init__(self, interface: Union["ensight_api.ensight", "ensight"]):
        self._ensight = interface

    @staticmethod
    def scoped_name(obj: Any) -> ContextManager:
        """Allow for the use of ``with`` to shorten APIs.

        In the EnSight and PyEnsight APIs, the interfaces can become lengthy.
        This class makes it possible to shorten APIs for modules, classes,
        and namespaces.

        Examples
        --------
        >>> sn = s.ensight.utils.support.scoped_name
        >>> with sn(s.ensight.objs.core) as core, sn(s.ensight.objs.enums) as enums:
        >>>     print(core.PARTS.find(True, enums.VISIBLE))
        >>> sn = ensight.utils.support.scoped_name
        >>> with sn(ensight.objs.core) as core, sn(ensight.objs.enums) as enums:
        >>>     print(core.PARTS.find(True, enums.VISIBLE))
        """
        return ScopedName(obj)


class ScopedName:
    """Allow for the use of ``with`` to shorten APIs.

    In the EnSight and PyEnsight APIs, the interfaces can become lengthy.
    This class makes it possible to shorten APIs for modules, classes,
    and namespaces.
    """

    def __init__(self, obj: Any):
        self._obj = obj

    def __enter__(self) -> Any:
        return self._obj

    def __exit__(self, exc_type, exc_value, exc_trace):
        pass



"""Docker Launcher module.

The Docker Launcher module provides PyEnSight with the ability to launch a
:class:`Session<ansys.pyensight.core.Session>` instance using a local Docker
installation or a `PIM <https://github.com/ansys/pypim`_-launched Docker installation.

Examples:
    ::

        from ansys.pyensight.core import DockerLauncher
        launcher = DockerLauncher(data_directory="D:\\data")
        launcher.pull()
        session = launcher.start()
        session.close()

"""
import logging
import os.path
import subprocess
import time
from typing import Any, Dict, Optional
import uuid

import urllib3

try:
    import grpc
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError("The grpc module must be installed for DockerLauncher")
except Exception:  # pragma: no cover
    raise RuntimeError("Cannot initialize grpc")

from ansys.pyensight.core.launcher import Launcher
import ansys.pyensight.core.session
from ansys.pyensight.core.session import Session

try:
    from ansys.pyensight.core import enshell_grpc
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError("The enshell_grpc must be installed for DockerLauncher")
except Exception:  # pragma: no cover
    raise RuntimeError("Cannot initialize enshell_grpc")


try:
    from simple_upload_server.client import Client

    simple_upload_server_is_available = True  # pragma: no cover
except Exception:
    simple_upload_server_is_available = False


class DockerLauncher(Launcher):
    """Creates a ``Session`` instance using a copy of EnSight hosted in a Docker container.

    This class allows you to either attach to a `PIM <https://github.com/ansys/pypim>`_-launched
    Docker container hosting EnSight or launch a local Docker container hosting EnSight.
    gRPC connections to EnShell and EnSight are established. A PyEnSight ``Session`` instance
    is returned upon successful launch and connection.

    Parameters
    ----------
    data_directory : str, optional
        Host directory to make into the container at ``/data``.
        The default is ``None``.
    docker_image_name : str, optional
        Name of the Docker image to use. The default is ``None``.
    use_dev : bool, optional
        Whether to use the latest ``ensight_dev`` Docker image. The
        default is ``False``. However, this value is overridden if the
        name of a Docker image is specified for the ``docker_image_name``
        parameter.
    channel :
        Existing gRPC channel to a running ``EnShell`` instance provided by PIM.
    pim_instance :
        PyPIM instance if using PIM (internal). The default is ``None``.
    timeout : float, optional
        Number of seconds to try a gRPC connection before giving up.
        This parameter is defined on the parent ``Launcher`` class,
        where the default is ``120``.
    use_egl : bool, optional
        Whether to use EGL hardware for accelerated graphics. The platform
        must be able to support this hardware. This parameter is defined on
        the parent ``Launcher`` class, where the default is ``False``.
    use_sos : int, optional
        Number of EnSight servers to use for SOS (Server of Server) mode.
        This parameter is defined on the parent ``Launcher`` class, where
        the default is ``None``, in which case SOS mode is not used.

    Examples
    --------
    >>> from ansys.pyensight.core import DockerLauncher
    >>> launcher = DockerLauncher(data_directory="D:\\data")
    >>> launcher.pull()
    >>> session = launcher.start()
    >>> session.close()

    """

    def __init__(
        self,
        data_directory: Optional[str] = None,
        docker_image_name: Optional[str] = None,
        use_dev: bool = False,
        channel: Optional[grpc.Channel] = None,
        pim_instance: Optional[Any] = None,
        **kwargs,
    ) -> None:
        """Initialize DockerLauncher."""
        super().__init__(**kwargs)

        self._data_directory = data_directory
        self._enshell_grpc_channel = channel
        self._service_uris: Dict[Any, str] = {}
        self._image_name: Optional[str] = None
        self._docker_client: Optional[Any] = None
        self._container = None
        self._enshell: Optional[Any] = None
        self._pim_instance: Optional[Any] = pim_instance

        # EnSight session secret key
        self._secret_key: str = str(uuid.uuid1())
        # temporary directory
        # it's in the ephemeral container, so just use "ensight's"
        # home directory within the container
        self._session_directory: str = "/home/ensight"
        # the Ansys / EnSight version we found in the container
        # to be reassigned later
        self._ansys_version: Optional[str] = None
        # EnShell's log file
        self._enshell_log_file = None
        # pointer to the file service object if available
        self._pim_file_service = None

        logging.debug(
            f"DockerLauncher data_dir={self._data_directory}\n"
            + f"  image_name={self._image_name}\n"
            + f"  use_dev={use_dev}\n"
        )

        if self._enshell_grpc_channel and self._pim_instance:
            if not set(("grpc_private", "http", "ws")).issubset(self._pim_instance.services):
                raise RuntimeError(
                    "If channel is specified, the PIM instance must have a list of length 3 "
                    + "containing the appropriate service URIs. It does not."
                )
            self._service_host_port = {}
            # grab the URIs for the 3 required services passed in from PIM
            self._service_host_port["grpc_private"] = self._get_host_port(
                self._pim_instance.services["grpc_private"].uri
            )
            self._service_host_port["http"] = self._get_host_port(
                self._pim_instance.services["http"].uri
            )
            self._service_host_port["ws"] = self._get_host_port(
                self._pim_instance.services["ws"].uri
            )
            # for parity, add 'grpc' as a placeholder even though using PIM sets up the gRPC channel.
            # this isn't used in this situation.
            self._service_host_port["grpc"] = ("127.0.0.1", -1)
            # attach to the file service if available
            self._get_file_service()
            return

        # EnShell gRPC port, EnSight gRPC port, HTTP port, WSS port
        # skip 1999 as that internal to the container is used to the container for the VNC connection
        ports = self._find_unused_ports(4, avoid=[1999])
        if ports is None:
            raise RuntimeError("Unable to allocate local ports for EnSight session")
        self._service_host_port = {}
        self._service_host_port["grpc"] = ("127.0.0.1", ports[0])
        self._service_host_port["grpc_private"] = ("127.0.0.1", ports[1])
        self._service_host_port["http"] = ("127.0.0.1", ports[2])
        self._service_host_port["ws"] = ("127.0.0.1", ports[3])

        # get the optional user-specified image name
        # Note: the default name needs to change over time...  TODO
        self._image_name = "ghcr.io/ansys-internal/ensight"
        if use_dev:
            self._image_name = "ghcr.io/ansys-internal/ensight_dev"
        if docker_image_name:
            self._image_name = docker_image_name

        # Load up Docker from the user's environment
        try:
            import docker

            self._docker_client = docker.from_env()
        except ModuleNotFoundError:
            raise RuntimeError("The pyansys-docker module must be installed for DockerLauncher")
        except Exception:
            raise RuntimeError("Cannot initialize Docker")

    def ansys_version(self) -> Optional[str]:
        """Get the Ansys version (three-digit string) found in the Docker container.

        Returns
        -------
        str
            Ansys version or ``None`` if not found or not started.

        """
        return self._ansys_version

    def pull(self) -> None:
        """Pull the Docker image.

        Raises
        ------
        RuntimeError
            If the Docker image couldn't be pulled.

        """
        try:
            if self._docker_client is not None:
                self._docker_client.images.pull(self._image_name)
        except Exception:
            raise RuntimeError(f"Can't pull Docker image: {self._image_name}")

    def start(self) -> "Session":
        """Start EnShell by running a local Docker EnSight image.

        Once this method starts EnShell, it connects over gRPC to EnShell running
        in the Docker container. Once connected, EnShell launches a copy of EnSight
        and WSS in the container. It then creates and binds a
        :class:`Session<ansys.pyensight.core.Session>` instance to the created
        EnSight gRPC connection and returns this instance.

        Returns
        -------
        object
            PyEnSight ``Session`` object instance.

        Raises
        ------
        RuntimeError
            Variety of error conditions.

        """
        tmp_session = super().start()
        if tmp_session:
            return tmp_session

        # Launch the EnSight Docker container locally as a detached container
        # initially running EnShell over the first gRPC port. Then launch EnSight
        # and other apps.

        # Create the environmental variables
        local_env = os.environ.copy()
        local_env["ENSIGHT_SECURITY_TOKEN"] = self._secret_key
        local_env["WEBSOCKETSERVER_SECURITY_TOKEN"] = self._secret_key
        # If for some reason, the ENSIGHT_ANSYS_LAUNCH is set previously,
        # honor that value, otherwise set it to "pyensight".  This allows
        # for an environmental setup to set the value to something else
        # (e.g. their "app").
        if "ENSIGHT_ANSYS_LAUNCH" not in local_env:
            local_env["ENSIGHT_ANSYS_LAUNCH"] = "container"

        # Environment to pass into the container
        container_env = {
            "ENSIGHT_SECURITY_TOKEN": self._secret_key,
            "WEBSOCKETSERVER_SECURITY_TOKEN": self._secret_key,
            "ENSIGHT_SESSION_TEMPDIR": self._session_directory,
            "ANSYSLMD_LICENSE_FILE": os.environ["ANSYSLMD_LICENSE_FILE"],
            "ENSIGHT_ANSYS_LAUNCH": local_env["ENSIGHT_ANSYS_LAUNCH"],
        }
        if "ENSIGHT_ANSYS_APIP_CONFIG" in local_env:
            container_env["ENSIGHT_ANSYS_APIP_CONFIG"] = local_env["ENSIGHT_ANSYS_APIP_CONFIG"]

        # Ports to map between the host and the container
        # If we're here in the code, then we're not using PIM
        # and we're not really using URIs where the hostname
        # is anything other than 127.0.0.1, so, we only need
        # to grab the port numbers.
        grpc_port = self._service_host_port["grpc"][1]
        ports_to_map = {
            str(self._service_host_port["grpc"][1])
            + "/tcp": str(self._service_host_port["grpc"][1]),
            str(self._service_host_port["grpc_private"][1])
            + "/tcp": str(self._service_host_port["grpc_private"][1]),
            str(self._service_host_port["http"][1])
            + "/tcp": str(self._service_host_port["http"][1]),
            str(self._service_host_port["ws"][1]) + "/tcp": str(self._service_host_port["ws"][1]),
        }

        # The data directory to map into the container
        data_volume = None
        if self._data_directory:
            data_volume = {self._data_directory: {"bind": "/data", "mode": "rw"}}

        # Start the container in detached mode with EnShell as a
        # gRPC server as the command
        #
        enshell_cmd = "-app -v 3 -grpc_server " + str(grpc_port)

        try:
            import docker
        except ModuleNotFoundError:  # pragma: no cover
            raise RuntimeError("The pyansys-docker module must be installed for DockerLauncher")
        except Exception:  # pragma: no cover
            raise RuntimeError("Cannot initialize Docker")

        use_egl = self._use_egl()

        logging.debug("Starting Container...\n")
        if data_volume:
            if use_egl:
                if self._docker_client:
                    self._container = self._docker_client.containers.run(
                        self._image_name,
                        command=enshell_cmd,
                        volumes=data_volume,
                        environment=container_env,
                        device_requests=[
                            docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])
                        ],
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                    )
            else:
                logging.debug(f"Running container {self._image_name} with cmd {enshell_cmd}\n")
                logging.debug(f"ports to map: {ports_to_map}\n")
                if self._docker_client:
                    self._container = self._docker_client.containers.run(
                        self._image_name,
                        command=enshell_cmd,
                        volumes=data_volume,
                        environment=container_env,
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                    )
                logging.debug(f"_container = {str(self._container)}\n")
        else:
            if use_egl:
                if self._docker_client:
                    self._container = self._docker_client.containers.run(
                        self._image_name,
                        command=enshell_cmd,
                        environment=container_env,
                        device_requests=[
                            docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])
                        ],
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                    )
            else:
                logging.debug(f"Running container {self._image_name} with cmd {enshell_cmd}\n")
                logging.debug(f"ports to map: {ports_to_map}\n")
                if self._docker_client:
                    self._container = self._docker_client.containers.run(
                        self._image_name,
                        command=enshell_cmd,
                        environment=container_env,
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                    )
                # logging.debug(f"_container = {str(self._container)}\n")
        logging.debug("Container started.\n")
        return self.connect()

    def connect(self):
        """Create and bind a :class:`Session<ansys.pyensight.core.Session>` instance
        to the created EnSight gRPC connection started by EnShell.

        This method is an internal method.

        Returns
        -------
        obj
            :class:`Session<ansys.pyensight.core.Session>` instance

        Raises
        ------
        RuntimeError:
            Variety of error conditions.
        """
        #
        #
        # Start up the EnShell gRPC interface
        log_dir = "/data"
        if self._enshell_grpc_channel:  # pragma: no cover
            self._enshell = enshell_grpc.EnShellGRPC()
            self._enshell.connect_existing_channel(self._enshell_grpc_channel)
            log_dir = "/work"
        else:
            logging.debug(
                f"Connecting to EnShell over gRPC port: {self._service_host_port['grpc'][1]}...\n"
            )
            self._enshell = enshell_grpc.EnShellGRPC(port=self._service_host_port["grpc"][1])
            time_start = time.time()
            while time.time() - time_start < self._timeout:
                if self._enshell.is_connected():
                    break
                try:
                    self._enshell.connect(timeout=self._timeout)
                except OSError:
                    pass

        if not self._enshell.is_connected():
            self.stop()
            raise RuntimeError("Can't connect to EnShell over gRPC.")

        cmd = "set_no_reroute_log"
        ret = self._enshell.run_command(cmd)
        logging.debug(f"enshell cmd: {cmd} ret: {ret}\n")
        if ret[0] != 0:
            self.stop()
            raise RuntimeError(f"Error sending EnShell command: {cmd} ret: {ret}")

        files_to_try = [log_dir + "/enshell.log", "/home/ensight/enshell.log"]
        for f in files_to_try:
            cmd = "set_debug_log " + f
            ret = self._enshell.run_command(cmd)
            if ret[0] == 0:
                self._enshell_log_file = f
                break
            else:
                logging.debug(f"enshell error; cmd: {cmd} ret: {ret}\n")

        if self._enshell_log_file is not None:
            logging.debug(f"enshell log file {self._enshell_log_file}\n")

        cmd = "verbose 3"
        ret = self._enshell.run_command(cmd)
        logging.debug(f"enshell cmd: {cmd} ret: {ret}\n")
        if ret[0] != 0:
            self.stop()
            raise RuntimeError(f"Error sending EnShell command: {cmd} ret: {ret}")

        logging.debug("Connected to EnShell.  Getting CEI_HOME and Ansys version...\n")
        logging.debug(f"  _enshell: {self._enshell}\n\n")
        # Build up the command to run ensight via the EnShell gRPC interface

        self._cei_home = self._enshell.cei_home()
        self._ansys_version = self._enshell.ansys_version()
        print("CEI_HOME=", self._cei_home)
        print("Ansys Version=", self._ansys_version)

        logging.debug("Got them.  Starting EnSight...\n")

        use_egl = self._use_egl()

        # Run EnSight
        ensight_env_vars = None
        if use_egl:
            ensight_env_vars = "LD_PRELOAD=/usr/local/lib64/libGL.so.1:/usr/local/lib64/libEGL.so.1"

        ensight_args = "-batch -v 3"

        if use_egl:
            ensight_args += " -egl"

        if self._use_sos:
            ensight_args += " -sos -nservers " + str(int(self._use_sos))

        ensight_args += " -grpc_server " + str(self._service_host_port["grpc_private"][1])

        vnc_url = "vnc://%%3Frfb_port=1999%%26use_auth=0"
        ensight_args += " -vnc " + vnc_url

        logging.debug(f"Starting EnSight with args: {ensight_args}\n")
        ret = self._enshell.start_ensight(ensight_args, ensight_env_vars)
        if ret[0] != 0:
            self.stop()
            raise RuntimeError(f"Error starting EnSight with args: {ensight_args}")

        logging.debug("EnSight started.  Starting wss...\n")

        # Run websocketserver
        wss_cmd = "cpython /ansys_inc/v" + self._ansys_version + "/CEI/nexus"
        wss_cmd += self._ansys_version + "/nexus_launcher/websocketserver.py"
        wss_cmd += " --http_directory " + self._session_directory
        # http port
        wss_cmd += " --http_port " + str(self._service_host_port["http"][1])
        # vnc port
        wss_cmd += " --client_port 1999"

        if self._enable_rest_api:
            # grpc port
            wss_cmd += " --grpc_port " + str(self._service_host_port["grpc_private"][1])

        # EnVision sessions
        wss_cmd += " --local_session envision 5"
        # websocket port
        wss_cmd += " " + str(self._service_host_port["ws"][1])

        logging.debug(f"Starting WSS: {wss_cmd}\n")
        ret = self._enshell.start_other(wss_cmd)
        if ret[0] != 0:
            self.stop()
            raise RuntimeError(f"Error starting WSS: {wss_cmd}\n")

        logging.debug("wss started.  Making session...\n")

        # build the session instance
        # WARNING: assuming the host is the same for grpc_private, http, and ws
        # This may not be true in the future if using PIM.
        # revise Session to handle three different hosts if necessary.
        use_sos = False
        if self._use_sos:
            use_sos = True
        session = ansys.pyensight.core.session.Session(
            host=self._service_host_port["grpc_private"][0],
            grpc_port=self._service_host_port["grpc_private"][1],
            html_port=self._service_host_port["http"][1],
            ws_port=self._service_host_port["ws"][1],
            install_path=None,
            secret_key=self._secret_key,
            timeout=self._timeout,
            sos=use_sos,
            rest_api=self._enable_rest_api,
        )
        session.launcher = self
        self._sessions.append(session)

        logging.debug("Return session.\n")

        return session

    def stop(self) -> None:
        """Release any additional resources allocated during launching."""
        if self._enshell:
            if self._enshell.is_connected():
                try:
                    logging.debug("Stopping EnShell.\n")
                    self._enshell.stop_server()
                except Exception:
                    pass
                self._enshell = None
        #
        if self._container:
            try:
                logging.debug("Stopping the Docker Container.\n")
                self._container.stop()
            except Exception:
                pass
            try:
                logging.debug("Removing the Docker Container.\n")
                self._container.remove()
            except Exception:
                pass
            self._container = None

        if self._pim_instance is not None:
            logging.debug("Deleting the PIM instance.\n")
            self._pim_instance.delete()
            self._pim_instance = None
        super().stop()

    def _get_file_service(self) -> None:  # pragma: no cover
        if simple_upload_server_is_available is False:
            return
        if self._pim_instance is None:
            return

        if "http-simple-upload-server" in self._pim_instance.services:
            self._pim_file_service = Client(
                token="token",
                url=self._pim_instance.services["http-simple-upload-server"].uri,
                headers=self._pim_instance.services["http-simple-upload-server"].headers,
            )

    def file_service(self) -> Optional[Any]:
        """Get the PIM file service object if available."""
        return self._pim_file_service

    def _get_host_port(self, uri: str) -> tuple:
        parse_results = urllib3.util.parse_url(uri)
        return (parse_results.host, parse_results.port)

    def _is_system_egl_capable(self) -> bool:
        """Check if the system is EGL capable.

        Returns
        -------
        bool
            ``True`` if the system is EGL capable, ``False`` otherwise.
        """
        if self._is_windows():
            return False

        # FIXME: MFK, need to figure out how we'd do this
        # with a PIM managed system such as Ansys Lab
        if self._pim_instance is not None:
            return False

        try:  # pragma: no cover
            subprocess.check_output("nvidia-smi")
            return True
        except (subprocess.CalledProcessError, FileNotFoundError):
            return False

    def enshell_log_contents(self) -> Optional[str]:
        """Get the contents of the EnShell log if possible.

        Return
        -------
        str
           Contents of the log or ``None``.
        """
        if self._enshell_log_file is None:
            return None

        if self._container is not None:
            try:
                # docker containers allow copying from a container to the host
                # a file, files, or a directory.  it arrives as a tar file.
                # we're grabbing just the log file, but we still need to
                # extract it from the tar file and then put the contents
                # in the string we're returning
                from io import BytesIO
                import tarfile

                bits, stat = self._container.get_archive(self._enshell_log_file)
                file_obj = BytesIO()
                for chunk in bits:
                    file_obj.write(chunk)
                file_obj.seek(0)
                tar = tarfile.open(mode="r", fileobj=file_obj)
                member = tar.getmembers()
                text = tar.extractfile(member[0])
                s = text.read().decode("utf-8")
                text.close()
                return s
            except Exception as e:
                logging.debug(f"Error getting EnShell log: {e}\n")
                return None

        fs = self.file_service()  # pragma: no cover
        if fs is not None:  # pragma: no cover
            try:
                fs.download_file("enshell.log", ".")
                f = open("enshell.log")
                s = f.read()
                f.close()
                return s
            except Exception:
                return None
